/*

Coobird's Blocks - Block breaking Java game.

Copyright (c) 2008, Chris Kroells
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the author nor the names of its contributors may be
      used to endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/


package net.coobird.blocks;

import java.awt.*;

/**
 * The game engine for Blocks.
 * @author coobird
 *
 */
public class BlocksEngine implements Runnable
{
	private int width = 400;
	private int height = 400;
	
	private int[] blockType;
	private int[] blockX;
	private int[] blockY;
	private int[] blockDX;
	private int[] blockDY;
	private int blockWidth;
	private int blockHeight;
	private Image[] blockImage;
	
	private static final int BLOCK_GONE = -1;
	
	private static final int DEFAULT_BLOCK_WIDTH = 40;
	private static final int DEFAULT_BLOCK_HEIGHT = 20;

	// Variables for balls.
	private int[] ballState;
	private int[] ballX;
	private int[] ballY;
	private int[] ballDX;
	private int[] ballDY;
	private int ballWidth;
	private int ballRadius;
	
	// The ball modes.
	private static final int BALL_NONE = 0; 
	private static final int BALL_NORMAL = 1; 
	private static final int BALL_SUPER = 2; 
	private static final int BALL_ITEM_50PTS = 3; 
	private static final int BALL_ITEM_ELONGATE = 4; 
	private static final int BALL_ITEM_HEALTH = 5; 
	private static final int BALL_ITEM_NEWBALL = 6; 
	private static final int BALL_ITEM_NEWSUPERBALL = 7; 

	// Variables for paddles.
	private int paddleX;
	private int paddleY;
	private int paddleWidth;
	private int paddleHeight;
	
	private static final int MAX_PADDLE_WIDTH = 100;
	private static final int MIN_PADDLE_WIDTH = 40;
	private static final int DEFAULT_PADDLE_HEIGHT = 5;
	
	private int level;
	private int score;
	private int damage;
	private int maxDamage;
	
	private int numBalls = 0;
	private int blocksLeft = 0;
	
	/**
	 * Probability for a new ball or item to appear. This is set during
	 * the makeLevel() method.
	 */
	private double newBallProbability = 0d;
	/**
	 * Number of balls that are allowed. This is set during the
	 * makeLevel() method..
	 */
	private int maxAllowedBalls = 0;
	
	// Probability for the super ball or normal ball to appear.
	private static final double PROB_SUPER_BALL = 0.1;
	private static final double PROB_NORMAL_BALL = 0.3;
	
	private final int BLOCK_COLORS = 6;

	private Image bgImage = null;
	private final static Color transBlack = new Color(0, 0, 0, 128);
	private Polygon star;
	
	/**
	 * <code>running</code> determines whether the game engine thread
	 * should continue to run.
	 */
	private boolean running = false;
	
	/** Determines whether the game is in demo mode. */
	private boolean demo = true;
	/** Determines whether the game is paused. */
	private boolean paused = true;
	/** Determines whether the game is in game move mode. */
	private boolean gameover = false;
	/** Enable debugging information to appear. */
	private boolean debug = false;
	/**
	 * Notifies whether one cycle of game calculations is complete, and
	 * the new state is "ready for pick up" by the drawImage() method.
	 */
	public boolean canPickUp = false;

	// Used to gauge performance (determines frames/sec drawn)
	private long startTime = 0;
	private int fps = 0;
	private int fcount = 0;
	
	private Thread gameEngineThread = null;

	private final Color[] demoColor = {	new Color(0, 0, 255, 192),
										new Color(0, 255, 0, 192),
										new Color(255, 0, 0, 192),
										new Color(255, 255, 0, 192) };
	
	/**
	 * Initializes the Blocks game engine.
	 * @param width The width of the game. If the width is set to less
	 * than 100 pixels., the width is set to 100 pixels.
	 * @param height The height of the game. If the height is set to less
	 * than 100 pixels., the width is set to 100 pixels.
	 */
	public BlocksEngine(int width, int height)
	{
		// If game area is too small, enlarge it.
		if (width < 100)
			width = 100;
		if (height < 100)
			height = 100;
		
		this.width = width;
		this.height = height;
		
		init();
	}
	
	/**
	 * Initializes the game engine.
	 * Called from the constructor.
	 */
	private void init()
	{
		level = 0;
		damage = 0;
		maxDamage = 5;
		score = 0;
		paddleWidth = MIN_PADDLE_WIDTH;
		paddleHeight = DEFAULT_PADDLE_HEIGHT;
		
	    // Points required to draw the item star.
		star = new Polygon();
	    star.addPoint(0, -5);
	    star.addPoint(1, -1);
	    star.addPoint(5, -1);
	    star.addPoint(2, 1);
	    star.addPoint(3, 5);
	    star.addPoint(0, 2);
	    star.addPoint(-3, 5);
	    star.addPoint(-2, 1);
	    star.addPoint(-5, -1);
	    star.addPoint(-1, -1);
	    star.addPoint(0, -5);
		
		makeLevel();
		startEngine();
	}

	/**
	 * Starts the game engine.
	 */
	public void startEngine()
	{
		// If the game engine thread is not running, then begin.
		// Else, ignore the request to start.
		if (gameEngineThread == null || !gameEngineThread.isAlive())
		{
			gameEngineThread = new Thread(this);
			gameEngineThread.start();
		}
	}

	/**
	 * Stops the game engine.
	 */
	public void stopEngine()
	{
		running = false;
		
		try
		{
			gameEngineThread.join(1000);
		}
		catch (InterruptedException e)
		{
			e.printStackTrace(System.out);
		}
	}

	/**
	 * Generates the level.
	 */
	private void makeLevel()
	{
		gameover = false;
		damage = 0;
		
		blockWidth = DEFAULT_BLOCK_WIDTH / ((level / BLOCK_COLORS) + 1);
		blockHeight = DEFAULT_BLOCK_HEIGHT;
		int offsetX = (level / BLOCK_COLORS) * blockWidth / 2;
		int offsetY = DEFAULT_BLOCK_HEIGHT;
		
		maxAllowedBalls = level + 5;
		newBallProbability = 0.1;
		
		int rowsOfBlocks = (level % BLOCK_COLORS) + BLOCK_COLORS;
		int blocksPerRow = (width / blockWidth) / ((level / BLOCK_COLORS) + 1);
		int blockXSpacing = (width - (blockWidth * blocksPerRow)) / blocksPerRow;//(level / BLOCK_COLORS) * blockWidth;
		int numBlocks = rowsOfBlocks * blocksPerRow;
		
		int numMovingBlocks = BLOCK_COLORS - (level % BLOCK_COLORS);
		int totalBlocks = numBlocks + numMovingBlocks;
		
		// Initializes the blocks
		blockType = new int[totalBlocks];
		blockX = new int[totalBlocks];
		blockY = new int[totalBlocks];
		blockDX = new int[totalBlocks];
		blockDY = new int[totalBlocks];
		
		// Element counter for the blocks.
		int element = 0;
		
		// Temporary variable used in for-loop.
		int curBlockType = 0;
		int numThisRow = 0;
		
		// Create the map of blocks
		for (int i = 0; i < rowsOfBlocks; i++)
		{
			// Sets the type of block for this row.
			curBlockType = i % (BLOCK_COLORS * 2);
			
			// Calculates the number of blocks in this row.
			numThisRow = blocksPerRow;
			
			// In even levels (above level 6):
			// If row is even, then show all blocks
			// If row is odd, then hide one block at the end
			if ( (level % 2 == 0) && level >= BLOCK_COLORS )
			{
				if (i % 2 == 0)
					numThisRow = blocksPerRow;
				else
					numThisRow = blocksPerRow - 1;
			}
			
			// Positions the blocks
			for (int j = 0; j < numThisRow; j++)
			{
				blockType[element] = curBlockType;
				
	
				if (level > 0 && level % 2 == 0 && i % 2 == 0)
					blockX[element] = (j * (blockWidth + blockXSpacing)) + offsetX;
				else if (level > 0 && level % 2 == 1)
					blockX[element] = (j * (blockWidth + blockXSpacing)) + offsetX;
				else if (level >= 6)
					blockX[element] = (j * (blockWidth + blockXSpacing)) + offsetX + (blockXSpacing / 2) + (blockWidth / 2);
				else 
					blockX[element] = (j * (blockWidth + blockXSpacing));
				
				blockY[element] = (i * blockHeight) + offsetY;
				blockDX[element] = 0;
				blockDY[element] = 0;
				element++;
			}
		}
		
		// Set up flying blocks
		for (int i = 0; i < numMovingBlocks - 2; i++)
		{
			if (i % 2 == 0)
			{
				blockType[element] = 1;
				blockX[element] = 0;
				blockY[element] = (rowsOfBlocks + i) * blockHeight + offsetY;
				blockDX[element] = 5;
				blockDY[element] = 0;
			}
			else
			{
				blockType[element] = 1;
				blockX[element] = width - blockWidth;
				blockY[element] = (rowsOfBlocks + i) * blockHeight + offsetY;
				blockDX[element] = -5;
				blockDY[element] = 0;
			}
			
			element++;
		}
		
		// Any leftover (over-allocated) blocks are hidden.
		for (; element < totalBlocks; element++)
			blockType[element] = BLOCK_GONE;
		
		// Initializes the balls
		ballState = new int[maxAllowedBalls];
		ballX = new int[maxAllowedBalls];
		ballY = new int[maxAllowedBalls];
		ballDX = new int[maxAllowedBalls];
		ballDY = new int[maxAllowedBalls];
		ballWidth = 10;
		ballRadius = ballWidth / 2;
		
		// Set paddle position and size.
		paddleX = 0;
		paddleY = height - 50;
		
		// Initializes all balls to not appear.
		for (int i = 0; i < maxAllowedBalls; i++)
			ballState[i] = BALL_NONE;
		
		// One ball will appear at beginning of level.
		ballState[0] = BALL_NORMAL;
		ballX[0] = width / 2;
		ballY[0] = offsetY + ((rowsOfBlocks + (numMovingBlocks - 2) + 1) * blockHeight);
		ballDX[0] = 0;
		ballDY[0] = 5;
	}

	/**
	 * Sets the level of the game.
	 * @param level Level of the game.
	 */
	public void setLevel(int level)
	{
		this.level = level;
		makeLevel();
	}
	
	/**
	 * Sets whether to put the game engine into demo mode.
	 * @param b Whether the game engine should be in demo mode.
	 */
	public void setDemo(boolean b)
	{
		this.demo = b;
		
		if (b)
		{
			setLevel(0);
		}
		else
			;
	}
	
	/**
	 * Returns whether the game engine is in demo mode.
	 * @return <code>true</code> if the game engine is in demo mode,
	 * <code>false</code> otherwise.
	 */
	public boolean isDemo()
	{
		return demo;
	}
	
	/**
	 * Ends the demo mode.
	 */
	public void endDemo()
	{
		setDemo(false);
		setLevel(0);
		setPause(true);
		score = 0;
		damage = 0;
		level = 0;
		paddleWidth = MIN_PADDLE_WIDTH;
	}
	
	/**
	 * Pauses the game engine.
	 * @param p Whether to pause the game engine.
	 */
	public void setPause(boolean p)
	{
		if (paused && gameover)
		{
			gameover = false;
			setDemo(true);
		}
		
		paused = p;
	}

	/**
	 * Returns whether the game engine is paused.
	 * @return True if paused.
	 */
	public boolean isPaused()
	{
		return paused;
	}
	
	/**
	 * Sets whether to put the engine in debug mode.
	 * @param b Whether to put engine in debug mode.
	 */
	public void setDebug(boolean b)
	{
		debug = b;
	}
	
	/**
	 * Returns the debug mode state of the game engine.
	 * @return <code>true</code> if the game engine is in debug mode,
	 * <code>false</code> if not.
	 */
	public boolean isDebug()
	{
		return debug;
	}

	/**
	 * Draws the game to a Graphics object.
	 * @param g Graphics object to draw to.
	 */	
	public void drawImage(Graphics g, boolean force)
	{
		if (g == null)
			return;
		
		if (!canPickUp && !force)
			return;
		
		if (!force)
			canPickUp = false;
		
		if (debug)
			fcount++;
		
		// Draw the background picture, if available
		// Else, paint a block
		if (bgImage != null)
		{
			g.drawImage(bgImage, 0, 0, null);
		}
		else
		{
			g.setColor(Color.blue);
			g.fillRect(0, 0, width, height);
		}
		
		// Draw the blocks
		g.setColor(Color.black);
		
		// TODO
		// This if statement is TEMPORARY
		// Fix the NullPointerException that occurs due to the blockImage
		// not initializing in time!!
		if (blockImage != null)
		{
			for (int i = 0; i < blockX.length; i++)
			{
				// If block is gone, then continue to next block
				// Else, draw the block of appropriate image
				if (blockType[i] == -1)
					continue;
				else
					g.drawImage(blockImage[blockType[i]], blockX[i], blockY[i], blockWidth, blockHeight, null);
			}
		}

		// Draw the balls
		for (int i = 0; i < ballX.length; i++)
		{
			// Draw different types of balls
			if (ballState[i] == BALL_NORMAL)
			{
				g.setColor(Color.black);
				g.fillOval(ballX[i] - ballRadius, ballY[i] - ballRadius, ballWidth, ballWidth);
			}
			else if (ballState[i] == BALL_SUPER)
			{
				g.setColor(Color.red);
				g.fillOval(ballX[i] - ballRadius, ballY[i] - ballRadius, ballWidth, ballWidth);
			}
			else if (ballState[i] > BALL_SUPER)
			{
				if (ballState[i] == BALL_ITEM_50PTS)
					g.setColor(Color.yellow);
				else if (ballState[i] == BALL_ITEM_ELONGATE)
					g.setColor(Color.orange);
				else if (ballState[i] == BALL_ITEM_HEALTH)
					g.setColor(Color.blue);
				else if (ballState[i] == BALL_ITEM_NEWBALL)
					g.setColor(Color.black);
				else if (ballState[i] == BALL_ITEM_NEWSUPERBALL)
					g.setColor(Color.red);
				
				star.translate(ballX[i], ballY[i]);
				g.fillPolygon(star);
				star.translate(-ballX[i], -ballY[i]);
			}
			else
				;
		}
		
		// Draw the paddle
		g.setColor(Color.yellow);
		g.fillRect(paddleX, paddleY, paddleWidth, paddleHeight);
		
		// Draw the score
		g.setColor(Color.black);
		g.drawString("Level: " + level, 16, 16);
		g.drawString("Score: " + score, 116, 16);
		g.drawString("Damage: " + damage + " / " + maxDamage, 216, 16);
		g.setColor(Color.white);
		g.drawString("Level: " + level, 15, 15);
		g.drawString("Score: " + score, 115, 15);
		g.drawString("Damage: " + damage + " / " + maxDamage, 215, 15);
		
		// Debug information
		if (debug)
		{
			if (System.currentTimeMillis() - startTime >= 1000)
			{
				fps = fcount;
				fcount = 0;
				startTime = System.currentTimeMillis();
			}
			
			g.setColor(transBlack);
			g.fillRect(0, 250, 100, 150);
			g.setColor(Color.white);
			g.drawString("fps: " + fps, 10, 270);
			g.drawString("level: " + level, 10, 285);
			g.drawString("score: " + score, 10, 300);
			g.drawString("damage: " + damage, 10, 315);
			g.drawString("x: " + ballX[0] + "  y: " + ballY[0], 10, 330);
			g.drawString("dx: " + ballDX[0] + "  dy: " + ballDY[0], 10, 345);
			g.drawString("numBalls: " + numBalls, 10, 360);
			g.drawString("blocksLeft: " + blocksLeft, 10, 375);
		}

		if (demo)
		{
			g.setColor(demoColor[0]);
			g.fillRect(2 * width / 6, 2 * height / 6, width / 6 + 1, height / 6 + 1);
			g.setColor(demoColor[1]);
			g.fillRect(3 * width / 6, 2 * height / 6, width / 6 + 1, height / 6 + 1);
			g.setColor(demoColor[2]);
			g.fillRect(2 * width / 6, 3 * height / 6, width / 6 + 1, height / 6 + 1);
			g.setColor(demoColor[3]);
			g.fillRect(3 * width / 6, 3 * height / 6, width / 6 + 1, height / 6 + 1);

			int stringWidth = g.getFontMetrics().stringWidth("Coobird's Blocks");
			int center = (width / 2) - (stringWidth / 2);

			g.setColor(Color.black);
			g.drawString("Coobird's Blocks", center, height / 2);
			g.setColor(Color.white);
			g.drawString("Coobird's Blocks", center - 1, height / 2 - 1);

			stringWidth = g.getFontMetrics().stringWidth("coobird.net");
			center = (width / 2) - (stringWidth / 2);
			
			g.setColor(Color.black);
			g.drawString("coobird.net", center, 2 * height / 3);
			g.setColor(Color.white);
			g.drawString("coobird.net", center - 1, 2 * height / 3 - 1);
		}

		// If the game is paused, then darken screen and show the
		// "Paused" message.
		if (paused || gameover)
		{
			g.setColor(transBlack);
			g.fillRect(0, 0, width, height);
			g.setColor(Color.white);
			String message = null;
			
			if (gameover)
				message = "Game Over";
			else if (paused)
				message = "Paused";
			
			int stringWidth = g.getFontMetrics().stringWidth(message);
			int center = (width / 2) - (stringWidth / 2);
			
			g.drawString(message, center, height / 2);
		}
		
		// The latest update is picked up, so need a new update.
		canPickUp = false;
		
		return;
	}
	
	/**
	 * Main thread for the game engine.
	 */
	public void run()
	{
		// Used for checking how long one cycle of thread takes.
		long startTime = 0;
		long elapsedTime = 0;
		
		// When running becomes false, the thread will die.
		running = true;
		
		// No update of engine has taken place.
		canPickUp = false;
		
		while (running)
		{
			// If paused, then skip the processing.
			if (paused)
			{
				try
				{
					Thread.sleep(50);
				}
				catch (Exception e)
				{
					e.printStackTrace(System.out);
				}
				continue;
			}
			
			startTime = System.currentTimeMillis();
			
			// Move the blocks
			for (int i = 0; i < blockX.length; i++)
			{
				// Skip moving block if it doesn't exist
				if (blockType[i] == BLOCK_GONE)
					continue;
				
				// Move the block
				blockX[i] += blockDX[i];
				blockY[i] += blockDY[i];

				// Bounce block off the side walls
				if (blockX[i] < 0 || (blockX[i] + blockWidth) > width)
					blockDX[i] = -blockDX[i];
			}

			// Reset ball counter
			numBalls = 0;
			
			// Move the balls
			for (int i = 0; i < maxAllowedBalls; i++)
			{
				// Ball is not in play, continue for-loop to next ball
				if (ballState[i] == BALL_NONE)
					continue;
				
				// Move the ball
				ballX[i] += ballDX[i];
				ballY[i] += ballDY[i];

				// Process ground hit for balls
				if ((ballY[i] + ballRadius) > height)
				{
					// Super ball loses ability when it hits the ground,
					// All items disappear when they hit the ground.
					if (ballState[i] == BALL_SUPER)
					{
						ballState[i] = BALL_NORMAL;
						break;
					}
					else if (ballState[i] > BALL_SUPER)
					{
						ballState[i] = BALL_NONE;
						break;
					}
					
					// Process damage
					damage++;
					if (damage == maxDamage)
					{
						ballState[i] = BALL_NONE;
						damage = 0;
						paddleWidth -= 10;
						if (paddleWidth < MIN_PADDLE_WIDTH)
							paddleWidth = MIN_PADDLE_WIDTH;
					}
				}
				
				// Bounce ball back if it hits the sides of the wall
				if ((ballX[i] - ballRadius) < 0 || (ballX[i] + ballRadius) > width)
				{
					if ((ballX[i] - ballRadius) < 0)
						ballX[i] = ballRadius;
					else
						ballX[i] = width - ballRadius;	
					
					ballDX[i] = -ballDX[i];
				}
				// Bounce off the ground or ceiling
				if ((ballY[i] - ballRadius) < 0 || (ballY[i] + ballRadius) > height)
				{
					if ((ballY[i] - ballRadius) < 0)
						ballY[i] = ballRadius;
					else
						ballY[i] = height - ballRadius;
					
					ballDY[i] = -ballDY[i];
				}
			}
			
			// Count number of balls
			for (int i = 0; i < maxAllowedBalls; i++)
				if (ballState[i] != BALL_NONE)
					numBalls++;

			
			// Check for collision
			// Nested for-loop checks collision between each ball and environment
			for (int i = 0; i < maxAllowedBalls; i++)
			{
				// If ball doesn't exist, skip this ball, proceed to next.
				// Else if, ball is an item, then skip.
				if (ballState[i] == BALL_NONE)
					continue;
				
				
				// Demo mode, move paddle
				if (demo)
				{
					paddleX = ballX[i] - (paddleWidth * (int)(1 / 3.0));
					
					if (paddleX < 0)
						paddleX = 0;
					else if (paddleX + paddleWidth > width)
						paddleX = width - paddleWidth;
				}
				
				// Collision check between ball and block
				for (int j = 0; j < blockX.length; j++)
				{
					// If the block doesn't exist, continue to next block
					if (blockType[j] == BLOCK_GONE)
						continue;

					// If ball is an item, then break
					if (ballState[i] > BALL_SUPER)
						break;
					
					// Check for collision between *ball and block*
					if ( (ballX[i] + ballRadius >= blockX[j]) && (ballX[i] - ballRadius <= blockX[j] + blockWidth) &&
						 (ballY[i] + ballRadius >= blockY[j]) && (ballY[i] - ballRadius <= blockY[j] + blockHeight) )
					{
						// Block is erased or degraded.
						if (blockType[j] >= BLOCK_COLORS)
							blockType[j] -= BLOCK_COLORS;
						else
							blockType[j] = BLOCK_GONE;
						
						// Increase the score
						score += 10;
						
						// Ball is super ball, so its movement is unaffected by blocks
						if (ballState[i] == BALL_SUPER)
							break;
						
						// Bounce the ball off block
						ballDX[i] = -ballDX[i];
						ballDY[i] = -ballDY[i];
						
						// if block was moving, transfer some momentum
						if (blockDX[j] > ballDX[i])
							ballDX[i] += (blockDX[j] / 2);
						
						// Check if block hit from side
						// (Undo the vertical reflection performed earlier)
						if ( (ballX[i] - ballRadius <= blockX[j]) || (ballX[i] + ballRadius >= blockX[j] + blockWidth) )
						{
							// undo reflection
							ballDY[i] = -ballDY[i];
							
							if (ballX[i] - ballRadius <= blockX[j])
							{
								ballDX[i] -= (blockX[j] - ballX[i] + ballRadius) * (1.0 / blockWidth);
								ballX[i] = blockX[j] - ballRadius;
							}
							else if (ballX[i] + ballRadius >= blockX[j] + blockWidth)
							{
								ballDX[i] += (ballX[i] + ballRadius - blockX[j]) * (1.0 / blockWidth);
								ballX[i] = blockX[j] + blockWidth + ballRadius;
							}
						}
						
						// Generates a new ball or item by random.
						if (Math.random() < newBallProbability)
						{
							double anotherRand = Math.random();
							int type = BALL_NONE;
							
							if (anotherRand < PROB_SUPER_BALL)
								type = BALL_SUPER;
							else if (anotherRand < PROB_NORMAL_BALL)
								type = BALL_NORMAL;
							else
							{
								int choice = (int)(Math.random() * 4);
								
								if (choice == 0)
									type = BALL_ITEM_50PTS;
								else if (choice == 1)
									type = BALL_ITEM_ELONGATE;
								else if (choice == 2)
									type = BALL_ITEM_HEALTH;
								else if (choice == 3)
									type = BALL_ITEM_NEWBALL;
								else if (choice == 4)
									type = BALL_ITEM_NEWSUPERBALL;
							}
							
							// If a ball, then make new ball at opposite direction
							// Else, it's an item, so just drop it.
							if (type <= BALL_SUPER)
								makeNewBall(ballX[i], ballY[i], -ballDX[i], -ballDY[i], type);
							else
								makeNewBall(ballX[i], ballY[i], 0, 5, type);
						}
						
						// Bounced off a block, so proceed
						break;
					}
				}
				
				// Check for collision between *ball and paddle*
				if ( (ballX[i] + ballRadius >= paddleX) && (ballX[i] - ballRadius < paddleX + paddleWidth) &&
						  (ballY[i] + ballRadius >= paddleY) && (ballY[i] - ballRadius < paddleY + paddleHeight) )
				{
					// Was this an item?
					// If so, process it and leave.
					if (ballState[i] > BALL_SUPER)
					{
						if (ballState[i] == BALL_ITEM_50PTS)
						{
							score += 50;
						}
						else if (ballState[i] == BALL_ITEM_ELONGATE)
						{
							if (paddleWidth < MAX_PADDLE_WIDTH)
								paddleWidth += 10;
						}
						else if (ballState[i] == BALL_ITEM_HEALTH)
						{
							damage = 0;
						}
						else if (ballState[i] == BALL_ITEM_NEWBALL)
						{
							makeNewBall(ballX[i], ballY[i], -ballDX[i], -ballDY[i], BALL_NORMAL);
						}
						
						// Item disappears.
						ballState[i] = BALL_NONE;
					}
					
					
					// Where on the paddle was the ball? (relative to center)
					int position = ballX[i] - (paddleX + (paddleWidth / 2));
					
					// If ball is "inside" the paddle, move it off
					if ((ballY[i] >= paddleY) && (ballY[i] <= paddleY + paddleHeight))
					{
						if (ballDY[i] > 0)
							ballY[i] = paddleY - ballRadius;
						else
							ballY[i] = paddleY + paddleHeight + ballRadius;
					}
						
					
					// Change the speed of ball depending on where on the paddle
					// the ball hit.
					ballDX[i] = ballDX[i] + (position / 4);
					ballDY[i] = -ballDY[i];					
				}
			}
			
			// Check if level complete.
			blocksLeft = 0;
			for (int i = 0; i < blockX.length; i++)
			{
				if (blockType[i] != BLOCK_GONE)
					blocksLeft++;
			}
			if (blocksLeft == 0)
			{
				if (demo && level > 100)
					level = 0;

				level++;
				makeLevel();
				
				// If a human player, then pause at new level.
				if (!demo)
					setPause(true);
			}
			
			if (numBalls == 0)
			{
				gameover = true;
				setPause(true);
			}
			
			elapsedTime = System.currentTimeMillis() - startTime;
			
			while (elapsedTime < 25)
			{
				// Delay
				try
				{
					Thread.sleep(1);
					elapsedTime = System.currentTimeMillis() - startTime;
				}
				catch (Exception e)
				{
					e.printStackTrace(System.out);
				}
			}
			
			// Finished processing. Update is ready for pick up.
			canPickUp = true;
		}
	}

	/**
	 * Makes a new ball.
	 * Internally, an item is classified as a ball.
	 * @param x The x position of the new ball.
	 * @param y The y position of the new ball.
	 * @param dx The x component of velocity of the new ball.
	 * @param dy The y component of velocity of the new ball.
	 * @param type The type for the new ball. (Refer to BlockEngine)
	 */
	private void makeNewBall(int x, int y, int dx, int dy, int type)
	{
		// Counter.
		int nextEmpty = 0;
		
		// Finds the next empty element in ball array
		for (nextEmpty = 0; nextEmpty < maxAllowedBalls; nextEmpty++)
			if (ballState[nextEmpty] == BALL_NONE)
				break;
		
		// All available elements in the ball array is filled.
		// Exits without making a new ball.
		if (nextEmpty == maxAllowedBalls)
			return;
		
		// Sets the new ball.
		ballState[nextEmpty] = type;
		ballX[nextEmpty] = x;
		ballY[nextEmpty] = y;
		ballDX[nextEmpty] = dx;
		ballDY[nextEmpty] = dy;
	}

	/**
	 * Sets the background image for the game.
	 * @param img A background image.
	 */
	public void setBackground(Image img)
	{
		this.bgImage = img;
	}

	/**
	 * Sets the images for the blocks.
	 * @param imgs An array of Image objects which contain the images
	 * used for the blocks
	 */
	public void setBlockImages(Image[] imgs)
	{
		blockImage = imgs;
	}

	/**
	 * Returns the width of a block.
	 * @return Width of a block.
	 */
	public int getBlockWidth()
	{
		return blockWidth;
	}

	/**
	 * Returns the height of a block.
	 * @return Height of a block.
	 */
	public int getBlockHeight()
	{
		return blockHeight;
	}

	/**
	 * Sets the location of the paddle in the game.
	 * @param x The horizontal location of the paddle.
	 * @param y The vertical location of the paddle.
	 */
	public void setPaddleLocation(int x, int y)
	{
		if (running && !demo)
		{
			if (x + paddleWidth > width)
				paddleX = width - paddleWidth;
			else
				paddleX = x;
		}
	}
	
	/**
	 * Moves the paddle by specified distance.
	 * 
	 */
	public void translatePaddle(int dx)
	{
		if (running && !demo)
		{
			// new X location
			int x = paddleX + dx;
			
			if (x + paddleWidth > width)
				paddleX = width - paddleWidth;
			else if (x < 0)
				paddleX = 0;
			else
				paddleX = x;
		}
	}
}