/*

Coobird's Blocks - Block breaking Java game.

Copyright (c) 2008, Chris Kroells
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the author nor the names of its contributors may be
      used to endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/


package net.coobird.blocks;

import java.awt.*;
import java.awt.event.*;

/**
 * An subclass of the Panel class, this is the main Panel which
 * displays the Blocks game screen and handles the user inputs.
 * @author coobird
 *
 */
public class BlocksPanel extends Panel
	implements MouseListener, MouseMotionListener, KeyListener
{
	private static final long serialVersionUID = -7723003358249718935L;

	private BlocksEngine engine;
	private BlocksPanelUpdate updater;
	
	// Image buffer
	private Image backImage;
	private Graphics backg;

	// The resolution used for the game.
	private final int WIDTH = 400;
	private final int HEIGHT = 400;
	
	private String backgroundImagePath = null;
	
	private boolean debug = false;
	
	private Thread updateThread;
	
	
	/**
	 * Default constructor.
	 */
	public BlocksPanel()
	{
		this.setPreferredSize(new Dimension(WIDTH, HEIGHT));
	}
		
	/**
	 * Initializes the BlocksPanel.
	 */
	public void init()
	{
		backImage = this.createImage(WIDTH, HEIGHT);
		backg = backImage.getGraphics();

		engine = new BlocksEngine(WIDTH, HEIGHT);
		engine.setBlockImages(initBlockImages());

		if (backgroundImagePath != null)
			engine.setBackground(initBackgroundImage(backgroundImagePath));
		else
			engine.setBackground(makeGradientBackground());
		
		updater = new BlocksPanelUpdate();
		
		this.addMouseListener(this);
		this.addMouseMotionListener(this);
		this.addKeyListener(this);

		engine.setDemo(true);
		engine.setPause(false);
		
		updateThread = new Thread(updater);
		updateThread.start();
	}
	
	/**
	 * Sets the external background image.
	 * @param filename Filename of the background image.
	 */
	public void setBackgroundImage(String filename)
	{
		backgroundImagePath = filename;
	}
	
	/**
	 * Sets the external background image.
	 * @param filename Filename of the background image.
	 */
	public void setBackgroundImage(Image img)
	{
		backgroundImagePath = null;
		
		if (img != null)
			engine.setBackground(img);
		else
			engine.setBackground(makeGradientBackground());
	}
	
	/**
	 * Makes a gradient background image.
	 * @return Background image.
	 */
	private Image makeGradientBackground()
	{
		Image bgImage = this.createImage(WIDTH, HEIGHT);
		
		// As reading background from file failed, make our own.
		Graphics g = bgImage.getGraphics();

		// Draws the gradient background
		float r0 = 0f;
		float g0 = 0f;
		float b0 = 64f;
		float step = 0.35f;
		
		for (int i = 0; i < HEIGHT; i++)
		{
			r0 += step;
			g0 += step;
			b0 += step;
			
			if (r0 > 240)
				r0 = 240;
			if (g0 > 240)
				g0 = 240;
			if (b0 > 240)
				b0 = 240;
			
			g.setColor(new Color((int)r0, (int)g0, (int)b0));
			g.drawLine(0, HEIGHT - i, i, HEIGHT);
		}
		for (int i = 0; i < HEIGHT; i++)
		{
			r0 += step;
			g0 += step;
			b0 += step;
			
			if (r0 > 240)
				r0 = 240;
			if (g0 > 240)
				g0 = 240;
			if (b0 > 240)
				b0 = 240;
			
			g.setColor(new Color((int)r0, (int)g0, (int)b0));
			g.drawLine(i, 0, WIDTH, HEIGHT - i);
		}
		
		g.dispose();
		g = null;
		
		return bgImage;
	}
	
	/**
	 * Loads a background image from specified location, and returns
	 * the image.
	 * If the image loading fails, then a gradient background is filled
	 * and returned.
	 * @param filename Image file to load as background.
	 * @return Background image.
	 */
	private Image initBackgroundImage(String filename)
	{
		// Create an Image
		Image bgImage = this.createImage(WIDTH, HEIGHT);
		
		System.out.println(filename);
		
		try
		{
			// Read image file for use as background.
			bgImage = Toolkit.getDefaultToolkit().createImage(filename);
		}
		catch (Exception e)
		{
			e.printStackTrace(System.out);
			
			bgImage = makeGradientBackground();
		}
		
		if (!new java.io.File(filename).exists())
			bgImage = makeGradientBackground();
		
		return bgImage;
	}
	
	/**
	 * Prepare images used for the blocks.
	 * @return An array of Image of the blocks.
	 */
	private Image[] initBlockImages()
	{
		// "12" corresponds to the types of blocks to make.
		// This is twice the value of BlocksEngine.BLOCK_COLORS.
		int numBlockTypes = 12;
		Image[] blockImage = new Image[numBlockTypes];
		
		// Gets the dimensions of the block to draw
		int blockWidth = engine.getBlockWidth();
		int blockHeight = engine.getBlockHeight();
		
		// Draw each brick type
		for (int k = 0; k < blockImage.length; k++)
		{
			blockImage[k] = this.createImage(blockWidth, blockHeight);
			Graphics g = blockImage[k].getGraphics();
			int step = 255 / blockHeight;
			
			int r0 = 0;
			int g0 = 0;
			int b0 = 0;
			
			switch (k % (numBlockTypes / 2))
			{
			case 0:
				r0 = 255;
				g0 = 255;
				b0 = 255;
				break;
			case 1:
				r0 = 255;
				g0 = 0;
				b0 = 0;
				break;
			case 2:
				r0 = 255;
				g0 = 0;
				b0 = 255;
				break;
			case 3:
				r0 = 0;
				g0 = 255;
				b0 = 255;
				break;
			case 4:
				r0 = 0;
				g0 = 255;
				b0 = 0;
				break;
			case 5:
				r0 = 255;
				g0 = 255;
				b0 = 0;
				break;
			default:
				r0 = 255;
				g0 = 255;
				b0 = 255;
			}
			
			// For any blockType > BlocksEngine.BLOCK_COLORS, the block
			// will be darkened.
			if (k >= (numBlockTypes / 2))
			{
				r0 /= 2;
				g0 /= 2;
				b0 /= 2;
				step /= 2;
			}
			
			// Draw the block
			for (int i = 0; i < blockHeight; i++)
			{
				r0 -= step;
				g0 -= step;
				b0 -= step;
				
				if (r0 < 0)
					r0 = 0;
				if (g0 < 0)
					g0 = 0;
				if (b0 < 0)
					b0 = 0;
				
				g.setColor(new Color(r0, g0, b0));
				g.drawLine(0, i, blockWidth, i);
			}
			
			g.dispose();
			g = null;
		}
		
		return blockImage;
	}
	
	/**
	 * Sets the level in the game engine.
	 * @param level The level.
	 */
	public void setLevel(int level)
	{
		engine.setLevel(level);
	}
	
	/**
	 * Paints the game display to the screen.
	 * @param g Graphics object for the Panel.
	 */
	public void paint(Graphics g)
	{
		if (engine != null && backg != null)
		{
			engine.drawImage(backg, false);
			g.drawImage(backImage, 0, 0, null);
		}
	}
	
	/**
	 * Draws the screen once.
	 * @param force Whether or not to force the game engine to draw
	 * the screen, even if the engine is in pause mode.
	 */
	private void drawScreen(boolean force)
	{
		if (!engine.canPickUp && !force)
			return;
		
		// Tells the GameEngine to draw to the backImage
		engine.drawImage(backg, force);

		// Creates a Graphics object from this Panel, and draws the
		// backImage buffer to the Panel.
		Graphics g = this.getGraphics();
		g.drawImage(backImage, 0, 0, null);
		
		g.dispose();
		g = null;
	}
	
	/**
	 * Stops the game engine and screen update threads.
	 */
	public void stop()
	{
		engine.stopEngine();
		
		updater.running = false;

		try
		{
			updateThread.join(1000);
		}
		catch (InterruptedException e)
		{
			e.printStackTrace(System.out);
		}
	}

	/**
	 * Handle a mouse movement.
	 * @param e MouseEvent from this event.
	 */
	public void mouseMoved(MouseEvent e)
	{
		if (!engine.isPaused())
			engine.setPaddleLocation(e.getX(), e.getY());
	}

	/**
	 * Handle a mouse click.
	 * @param e MouseEvent from this event.
	 */
	public void mouseClicked(MouseEvent e)
	{
		// If left-clicked
		// Else if right-clicked
		if (e.getButton() == MouseEvent.BUTTON1)
		{
			if (engine.isDemo())
				engine.endDemo();
			else if (engine.isPaused())
				engine.setPause(false);
		}
		else if (e.getButton() == MouseEvent.BUTTON3)
		{
			if (!engine.isDemo() && !engine.isPaused())
				engine.setPause(true);
		}
	}

	public void mouseDragged(MouseEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	public void mousePressed(MouseEvent e) {}
	public void mouseReleased(MouseEvent e) {}
	public void keyReleased(KeyEvent e)	{}
	public void keyTyped(KeyEvent e) {}

	/**
	 * Process a key press.
	 * @param e KeyEvent.
	 */
	public void keyPressed(KeyEvent e)
	{
		int keycode = e.getKeyCode();
		int modifier = e.getModifiers();
		
		if (keycode == KeyEvent.VK_D)
		{
			engine.setDebug(!engine.isDebug());
		}
		else if (keycode == KeyEvent.VK_P)
		{
			if (!engine.isDemo())
				engine.setPause(!engine.isPaused());
		}
		else if (keycode == KeyEvent.VK_F2)
		{
			if (!engine.isDemo())
				engine.setDemo(true);
		}
		else if (keycode == KeyEvent.VK_LEFT)
		{
			if (modifier == InputEvent.SHIFT_MASK)
				engine.translatePaddle(-10);
			else
				engine.translatePaddle(-5);
		}
		else if (keycode == KeyEvent.VK_RIGHT)
		{
			if (modifier == InputEvent.SHIFT_MASK)
				engine.translatePaddle(10);
			else
				engine.translatePaddle(5);
		}
	}

	/**
	 * A thread that runs and calls the game engine to update
	 * the image on the Panel.
	 * @author coobird
	 */
	class BlocksPanelUpdate implements Runnable
	{
		public boolean running = true;
		private boolean enginePaused = true;
		private long startTime = 0;
		private int fps = 0;
		
		public void run()
		{
			if (debug)
				startTime = System.currentTimeMillis();
			
			while (running)
			{
				/*
					enginePaused is used to determine when to call the
					drawScreen(boolean) method.
					It is toggled when the engine.isPaused() changes.

					The main reason this mechanism is used is to reduce
					the number of calls to drawScreen(boolean) method as
					it is the most resource intensive operation in Blocks.
					
					When the game engine is paused, there will be one call
					to drawScreen(true), which will show the "Paused" message
					and then, there will not be any refreshes.
				*/
				
				if (!engine.isPaused())
				{
					drawScreen(false);
					enginePaused = false;
					fps++;
				}
				else if (engine.isPaused() && !enginePaused)
				{
					drawScreen(true);
					enginePaused = true;
				}
				else
					;
				
				try
				{
					Thread.sleep(10);
				}
				catch (Exception e)
				{
					e.printStackTrace(System.out);
				}
				
				if (debug)
				{
					if ((System.currentTimeMillis() - startTime) > 1000)
					{
						System.out.println("fps: " + fps);
						startTime = System.currentTimeMillis();
						fps = 0;
					}
				}
			}
		}
	}
}
